@@ -8,9 +8,10 @@ from enum import Enum
 from .baseSIMulator import import_cp, import_torch, torch_GPU
 from .hexSIMulator import HexSim_simulator, RightHexSim_simulator
 from .conSIMulator import ConSim_simulator
+from .Illumination import ConIll, HexIll, RaHexIll
 from qtpy.QtWidgets import QWidget, QVBoxLayout, QFileDialog
 from napari.qt.threading import thread_worker
-from magicgui.widgets import SpinBox, Label, Container, ComboBox, FloatSpinBox, LineEdit
+from magicgui.widgets import SpinBox, Label, Container, ComboBox, FloatSpinBox, LineEdit, RadioButtons, PushButton
 from napari.layers import Layer
 import tifffile
 import numpy as np
@@ -18,7 +19,7 @@ import open3d as o3d
 
 
 class Samples(Enum):
-    POINTS = 500
+    SPHEROID = 500
     FILAMENTS = 10
 
 
@@ -43,11 +44,11 @@ class PointCloud(QWidget):
         self._viewer.layers.events.removed.connect(function.reset_choices)
         _layout.addWidget(function.native)
 
-    def point_cloud_fil(self, nSamples):
+    def point_cloud_fil(self, nSamples, L, dL):
         """Generates a point-cloud as the object in the imaging system."""
 
-        L = 5  # full length of each filament in um
-        dL = 0.05  # target length of each short bit to form a filament
+        # L: full length of each filament in um
+        # dL: target length of each short bit to form a filament
 
         for i in range(nSamples):
             alpha = 2 * np.pi * np.random.rand()
@@ -94,57 +95,77 @@ class PointCloud(QWidget):
                 points = np.single(np.dstack([x, y, z]).squeeze())
             else:
                 points = np.concatenate((points, np.single(np.dstack([x, y, z]).squeeze())))
-        npoints = points.shape[0]
         return points
 
-    def point_cloud(self, nSamples):
+    def point_cloud(self, npoints, rad, dep):
         """Generates a point - cloud as the object in the imaging system."""
 
-        rad = 5  # radius of sphere of points
+        # rad: radius of sphere of points
         # Multiply the points several times to get the enough number
-        pointsxn = (2 * np.random.rand(nSamples * 3, 3) - 1) * [rad, rad, rad]
+        pointsxn = (2 * np.random.rand(npoints * 3, 3) - 1) * [rad, rad, rad]
 
         pointsxnr = np.sum(pointsxn * pointsxn, axis=1)  # multiple times the points
         points_sphere = pointsxn[pointsxnr < (rad ** 2), :]  # simulate spheres from cubes
-        points = np.single(points_sphere[(range(nSamples)), :])
-        points[:, 2] = points[:, 2] / 2  # to make the point cloud for OTF a ellipsoid rather than a sphere
-        npoints = nSamples
+        points = np.single(points_sphere[(range(npoints)), :])
+        points[:, 2] = dep / rad * points[:, 2]  # to make the point cloud for OTF a ellipsoid rather than a sphere
         return points
 
-    def gen_point_cloud(self):
+    def gen_pc(self):
         if self.w_samples.value == Samples.FILAMENTS:
-            self.pc = self.point_cloud_fil(self.w_nSample.value)
-        elif self.w_samples.value == Samples.POINTS:
-            self.pc = self.point_cloud(self.w_nSample.value)
+            self.pc = self.point_cloud_fil(self.fil_n.value, self.fil_len.value, self.fil_step.value)
+        elif self.w_samples.value == Samples.SPHEROID:
+            self.pc = self.point_cloud(self.sph_points.value, self.sph_rad.value, self.sph_dep.value)
         print('Point cloud generated')
         if hasattr(self, 'pc'):
             try:
-                self._viewer.add_points(self.pc, size=0.1, name=self.w_samples.value)
+                self._viewer.add_points(-self.pc[:, ::-1], size=self.fil_step.value, name=self.w_samples.value)
             except Exception as e:
                 print(e)
 
-    def save_point_cloud(self):
+    def save_pc(self):
         if hasattr(self._viewer.layers.selection.active, 'data'):
             try:
                 options = QFileDialog.Options()
-                filename = QFileDialog.getSaveFileName(self, "Pick a file", options=options, filter="Files (*.pcd)")
+                filename = QFileDialog.getSaveFileName(self, 'Save a file', options=options, filter='Files (*.pcd)')
                 pcd = o3d.geometry.PointCloud()
-                # pcd.points = o3d.utility.Vector3dVector(self.pc)
-                pcd.points = o3d.utility.Vector3dVector(self._viewer.layers.selection.active.data)
+                pcd.points = o3d.utility.Vector3dVector(-self._viewer.layers.selection.active.data[:, ::-1])
                 o3d.io.write_point_cloud(filename[0], pcd)
                 print('Point cloud saved')
             except Exception as e:
                 print(e)
 
+    def load_pc(self):
+        try:
+            options = QFileDialog.Options()
+            filename = QFileDialog.getOpenFileName(self, 'Pick a file', options=options, filter='Files (*.pcd)')
+            pcd = o3d.io.read_point_cloud(filename[0])
+            out_arr = np.asarray(pcd.points)
+            self._viewer.add_points(out_arr, size=0.1, name=filename[0])
+        except Exception as e:
+            print(e)
+
     def cloud_widgets(self):
         """Creates a widget containing all small widgets"""
-        self.w_samples = ComboBox(value=Samples.FILAMENTS, label='sample', choices=Samples)
-        self.w_nSample = SpinBox(value=2, name='spin', label='N_samples')
-        self.w_gen_and_save = Container(widgets=[magicgui(self.gen_point_cloud, call_button='Generate point cloud',
-                                        auto_call=False),
-                                        magicgui(self.save_point_cloud, call_button='Save current layer as .pcd',
-                                        auto_call=True)], layout="horizontal", labels=None)
-        self.c_w = Container(widgets=[self.w_samples, self.w_nSample, self.w_gen_and_save])
+        self.w_samples = RadioButtons(value=Samples.SPHEROID, choices=Samples)
+
+        self.sph_points = SpinBox(value=1000, step=50, label='spheroid_points', max=10000)
+        self.sph_rad = SpinBox(value=5, label='spheroid_radius (μm)')
+        self.sph_dep = FloatSpinBox(value=2.5, step=0.5, max=self.sph_rad.value, label='spheroid_depth (μm)')
+        self.w_sph = Container(widgets=[self.sph_points, self.sph_dep, self.sph_rad])
+
+        self.fil_n = SpinBox(value=20, max=100, label='filament_n')
+        self.fil_len = SpinBox(value=5, label='filament_length (μm)')
+        self.fil_step = FloatSpinBox(value=0.05, step=0.01, max=self.fil_len.value, label='filament_step (μm)')
+        self.w_fil = Container(widgets=[self.fil_n, self.fil_len, self.fil_step])
+
+        self.comprehensive_w = Container(widgets=[magicgui(self.save_pc, call_button='Save current layer as .pcd',
+                                                           auto_call=False),
+                                                  magicgui(self.load_pc, call_button='Load point cloud',
+                                                           auto_call=False)], layout="horizontal", labels=None)
+        self.c_w = Container(widgets=[self.w_samples, self.w_sph, self.w_fil, magicgui(self.gen_pc,
+                                                                                       call_button='Generate point cloud',
+                                                                                       auto_call=False),
+                                      self.comprehensive_w], labels=None)
 
 
 class Sim_mode(Enum):
@@ -154,9 +175,11 @@ class Sim_mode(Enum):
 
 
 class Pol(Enum):
-    IN_PLANE = 0
-    AXIAL = 1
-    CIRCULAR = 2
+    azimuthal = 0  # in_plane
+    radial = 1  # axial
+    circular = 2
+    horizontal = 3
+    vertical = 4
 
 
 class Psf_calc(Enum):
@@ -166,12 +189,13 @@ class Psf_calc(Enum):
 
 class Accel(Enum):
     NUMPY = 0
+    if import_cp:
+        CUPY = 3
     if import_torch:
         TORCH_CPU = 1
         if torch_GPU:
             TORCH_GPU = 2
-    if import_cp:
-        CUPY = 3
+
 
 
 class SIMulator(QWidget):
@@ -195,7 +219,6 @@ class SIMulator(QWidget):
         self.wrap_widgets()
         layout = QVBoxLayout()
         self.setLayout(layout)
-        self.add_magic_function(self.select_layer, layout)
         self.add_magic_function(self.w, layout)
 
     def add_magic_function(self, function, _layout):
@@ -206,16 +229,16 @@ class SIMulator(QWidget):
 
     def parameters(self):
         self.SIM_mode = ComboBox(value=Sim_mode.SIM_CONV, label='SIM_mode', choices=Sim_mode)
-        self.Polarisation = ComboBox(value=Pol.AXIAL, label='Polarisation', choices=Pol)
+        self.Polarisation = ComboBox(value=Pol.radial, label='Polarisation', choices=Pol)
         self.Acceleration = ComboBox(value=list(Accel)[-1], label='Acceleration', choices=Accel)
         self.Psf = ComboBox(value=Psf_calc.SCALAR, label='Psf calculation', choices=Psf_calc)
         self.N = SpinBox(value=128, name='spin', label='N pixel')
         self.pixel_size = FloatSpinBox(value=6.5, name='spin', label='pixel size(μm)', step=0.5)
         self.magnification = SpinBox(value=60, name='spin', label='magnification')
-        self.ill_NA = FloatSpinBox(value=1.0, name='spin', label='NA  illumination', min=0.0, step=0.1)
+        self.ill_NA = FloatSpinBox(value=1, name='spin', label='NA  illumination', min=0.0, step=0.1)
         self.det_NA = FloatSpinBox(value=1.0, name='spin', label='NA  detection', min=0.0, step=0.1)
         self.n = FloatSpinBox(value=1.33, name='spin', label='n', min=0.00)
-        self.ill_wavelength = SpinBox(value=500, label='λ  illumination(nm)', step=50)
+        self.ill_wavelength = SpinBox(value=540, label='λ  illumination(nm)', step=50)
         self.det_wavelength = SpinBox(value=540, label='λ  detection(nm)', step=50)
 
         self.zrange = FloatSpinBox(value=3.5, name='spin', label='z range(μm)', min=0.0)
@@ -240,21 +263,27 @@ class SIMulator(QWidget):
     def set_att(self):
         """Sets attributes in the simulation class. Executed frequently to update the parameters"""
         if self.SIM_mode.value == Sim_mode.HEXSIM:
-            self.sim = HexSim_simulator()
+            # self.sim = HexSim_simulator()
+            self.sim = HexIll()
             nsteps = self.sim._phaseStep * self.sim._angleStep
         elif self.SIM_mode.value == Sim_mode.HEXSIM_RA:
-            self.sim = RightHexSim_simulator()
+            self.sim = RaHexIll()
             nsteps = self.sim._phaseStep * self.sim._angleStep
         elif self.SIM_mode.value == Sim_mode.SIM_CONV:
-            self.sim = ConSim_simulator()
+            # self.sim = ConSim_simulator()
+            self.sim = ConIll()
             nsteps = self.sim._phaseStep * self.sim._angleStep
 
-        if self.Polarisation.value == Pol.IN_PLANE:
-            self.sim.pol = 'in-plane'
-        elif self.Polarisation.value == Pol.AXIAL:
-            self.sim.pol = 'axial'
-        elif self.Polarisation.value == Pol.CIRCULAR:
-            self.sim.pol = 'circular'
+        if self.Polarisation.value == Pol.azimuthal:
+            self.sim.pol = 'a'
+        elif self.Polarisation.value == Pol.radial:
+            self.sim.pol = 'r'
+        elif self.Polarisation.value == Pol.circular:
+            self.sim.pol = 'c'
+        elif self.Polarisation.value == Pol.horizontal:
+            self.sim.pol = 'h'
+        elif self.Polarisation.value == Pol.vertical:
+            self.sim.pol = 'v'
 
         if self.Acceleration.value == Accel.NUMPY:
             self.sim.acc = 0
@@ -283,8 +312,8 @@ class SIMulator(QWidget):
         self.sim.ill_NA = self.ill_NA.value
         self.sim.det_NA = self.det_NA.value
         self.sim.n = self.n.value
-        self.sim.ill_wavelength = self.ill_wavelength.value
-        self.sim.det_wavelength = self.det_wavelength.value
+        self.sim.ill_wavelength = self.ill_wavelength.value * 1e-3
+        self.sim.det_wavelength = self.det_wavelength.value * 1e-3
         self.sim.zrange = self.zrange.value
         self.sim.tpoints = (self.tpoints.value // nsteps // 2) * nsteps * 2
         self.tpoints.value = self.sim.tpoints
@@ -317,7 +346,6 @@ class SIMulator(QWidget):
         if hasattr(self, 'sim'):
             delattr(self, 'sim')
 
-    @magicgui(call_button='Select image layer')
     def select_layer(self, layer: Layer):
         """
         Selects a layer used to simulate raw SIM stacks, it contains the raw point-cloud data.
@@ -327,9 +355,9 @@ class SIMulator(QWidget):
             return
         if hasattr(self, 'points'):
             delattr(self, 'points')
-        self.points = layer.data
+        self.points = -layer.data[:, ::-1]
         self.npoints = self.points.shape[0]
-        self.messageBox.value = f'Selected image layer: {layer.name}'
+        self.messageBox.value = f'Selected sample layer: {layer.name}'
 
     def get_results(self):
         if not hasattr(self, 'points'):
@@ -337,6 +365,12 @@ class SIMulator(QWidget):
         else:
             def show_img(data):
                 self._viewer.add_image(data, name='raw image stack',
+                                       scale=(self.zdrift.value * 0.001,
+                                              self.pixel_size.value / self.magnification.value,
+                                              self.pixel_size.value / self.magnification.value),
+                                       translate=(-self.zdrift.value * 0.001 * self.tpoints.value / 2,
+                                                  -self.pixel_size.value / self.magnification.value * self.N.value / 2,
+                                                  -self.pixel_size.value / self.magnification.value * self.N.value / 2),
                                        metadata={'mode': str(self.SIM_mode.value),
                                                  'pol': str(self.Polarisation.value),
                                                  'acc': str(self.Acceleration.value), 'psf': str(self.Psf.value),
@@ -373,65 +407,6 @@ class SIMulator(QWidget):
 
             _get_results()
 
-    def show_raw_img_sum(self, show_raw_img_sum: bool = False):
-        if hasattr(self, 'sim'):
-            if hasattr(self.sim, 'img_sum_z'):
-                if show_raw_img_sum:
-                    if self.used_par_list != self.par_list():
-                        self.messageBox.value = 'Parameters changed! Calculate the raw-image stack first!'
-                    else:
-                        try:
-                            self._viewer.add_image(self.sim.img_sum_z, name='raw image sum along z axis')
-                            self._viewer.add_image(self.sim.img_sum_x, name='raw image sum along x (or y) axis')
-                        except Exception as e:
-                            print(str(e))
-
-    def show_psf(self, show_3D_psf_slice: bool = False):
-        if hasattr(self, 'sim'):
-            if hasattr(self.sim, 'psf_z0'):
-                if show_3D_psf_slice:
-                    if self.used_par_list != self.par_list():
-                        self.messageBox.value = 'Parameters changed! Calculate the raw-image stack first!'
-                    else:
-                        try:
-                            self._viewer.add_image(self.sim.psf_z0, name='PSF in x-y plane')
-                        except Exception as e:
-                            print(e)
-
-    def show_otf(self, show_3D_otf_slice: bool = False):
-        if hasattr(self, 'sim'):
-            if hasattr(self.sim, 'aotf_x'):
-                if show_3D_otf_slice:
-                    if self.used_par_list != self.par_list():
-                        self.messageBox.value = 'Parameters changed! Calculate the raw-image stack first!'
-                    else:
-                        try:
-                            self._viewer.add_image(self.sim.aotf_x, name='OTF perpendicular to x')
-                        except Exception as e:
-                            print(str(e))
-
-    def save_tif_with_tags(self):
-        """Saves the selected image layer as a tif file with tags"""
-        if hasattr(self._viewer.layers.selection.active, 'data'):
-            try:
-                options = QFileDialog.Options()
-                filename = QFileDialog.getSaveFileName(self, "Pick a file", options=options, filter="Images (*.tif)")
-                tifffile.imwrite(filename[0], self._viewer.layers.selection.active.data,
-                                 description=str(self._viewer.layers.selection.active.metadata))
-            except Exception as e:
-                print(str(e))
-
-    def print_tif_tags(self):
-        """Prints tags of the selected tif image"""
-        try:
-            frames = tifffile.TiffFile(self._viewer.layers.selection.active.source.path)
-            page = frames.pages[0]
-            # Print file description
-            print(f'==={self._viewer.layers.selection.active.name}.tif===\n' + page.tags["ImageDescription"].value)
-            self.messageBox.value = 'Parameters printed'
-        except Exception as e:
-            print(str(e))
-
     def wrap_widgets(self):
         """Creates a widget containing all small widgets"""
         w_parameters = Container(
@@ -441,14 +416,130 @@ class SIMulator(QWidget):
                                    self.ill_wavelength, self.det_wavelength]),
                 Container(widgets=[self.magnification, self.zrange, self.tpoints, self.xdrift,
                                    self.zdrift, self.fwhmz, self.random_seed, self.drift, self.defocus,
-                                   self.sph_abb])], layout="horizontal")
-        w_cal = magicgui(self.get_results, call_button="Calculate raw image stack", auto_call=False)
-        w_save_and_print = Container(widgets=[magicgui(self.save_tif_with_tags, call_button='save_tif_with_tags'),
-                                              magicgui(self.print_tif_tags, call_button='print_tags')],
-                                     layout="horizontal", labels=None)
-        w_sum = magicgui(self.show_raw_img_sum, auto_call=True)
-        w_psf = magicgui(self.show_psf, auto_call=True)
-        w_otf = magicgui(self.show_otf, auto_call=True)
-        self.messageBox = LineEdit(value="Messages")
-        self.w = Container(widgets=[w_parameters, w_cal, w_save_and_print, w_sum, w_psf, w_otf, self.messageBox],
+                                   self.sph_abb])], layout='horizontal')
+        w_cal = magicgui(self.get_results, call_button='Calculate raw image stack', auto_call=False)
+
+        # 'save and print' widgets
+        save_tif_with_tags = PushButton(text='save_tif_with_tags')
+
+        @save_tif_with_tags.clicked.connect
+        def on_save_tif_with_tags_click():
+            """Saves the selected image layer as a tif file with tags"""
+            if hasattr(self._viewer.layers.selection.active, 'data'):
+                try:
+                    options = QFileDialog.Options()
+                    filename = QFileDialog.getSaveFileName(self, "Save a file", options=options,
+                                                           filter='Images (*.tif)')
+                    tifffile.imwrite(filename[0], self._viewer.layers.selection.active.data,
+                                     description=str(self._viewer.layers.selection.active.metadata))
+                except Exception as e:
+                    print(str(e))
+
+        show_raw_img_sum = PushButton(text='raw image sum')
+
+        @show_raw_img_sum.clicked.connect
+        def on_raw_image_sum_click():
+            if hasattr(self, 'sim'):
+                if hasattr(self.sim, 'img_sum_z'):
+                    if self.used_par_list != self.par_list():
+                        self.messageBox.value = 'Parameters changed! Calculate the raw-image stack first!'
+                    else:
+                        try:
+                            self._viewer.add_image(self.sim.img_sum_z,
+                                                   scale=(self.pixel_size.value / self.magnification.value,
+                                                          self.pixel_size.value / self.magnification.value),
+                                                   translate=(
+                                                       -self.pixel_size.value / self.magnification.value * self.N.value / 2,
+                                                       -self.pixel_size.value / self.magnification.value * self.N.value / 2),
+                                                   name='raw image sum along z axis')
+                            self._viewer.add_image(self.sim.img_sum_x,
+                                                   scale=(self.pixel_size.value / self.magnification.value,
+                                                          self.pixel_size.value / self.magnification.value),
+                                                   translate=(
+                                                       -self.pixel_size.value / self.magnification.value * self.N.value / 2,
+                                                       -self.pixel_size.value / self.magnification.value * self.N.value / 2),
+                                                   name='raw image sum along x (or y) axis')
+                        except Exception as e:
+                            print(str(e))
+
+        show_psf = PushButton(text='show_psf')
+
+        @show_psf.clicked.connect
+        def on_show_psf_click():
+            if hasattr(self, 'sim'):
+                if hasattr(self.sim, 'psf_z0'):
+                    if self.used_par_list != self.par_list():
+                        self.messageBox.value = 'Parameters changed! Calculate the raw-image stack first!'
+                    else:
+                        try:
+                            self._viewer.add_image(self.sim.psf_z0,
+                                                   scale=(self.pixel_size.value / self.magnification.value,
+                                                          self.pixel_size.value / self.magnification.value),
+                                                   translate=(
+                                                       -self.pixel_size.value / self.magnification.value * self.N.value / 2,
+                                                       -self.pixel_size.value / self.magnification.value * self.N.value / 2),
+                                                   name='PSF in x-y plane')
+                        except Exception as e:
+                            print(e)
+
+        print_tif = PushButton(text='print_tags')
+
+        @print_tif.clicked.connect
+        def on_print_tif_click():
+            """Prints tags of the selected tif image"""
+            try:
+                frames = tifffile.TiffFile(self._viewer.layers.selection.active.source.path)
+                page = frames.pages[0]
+                # Print file description
+                print(f'==={self._viewer.layers.selection.active.name}.tif===\n' + page.tags["ImageDescription"].value)
+                self.messageBox.value = 'Parameters printed'
+            except Exception as e:
+                print(str(e))
+
+        show_otf = PushButton(text='show_otf')
+        @show_otf.clicked.connect
+        def on_show_otf_click():
+            if hasattr(self, 'sim'):
+                if hasattr(self.sim, 'aotf_x'):
+                    if self.used_par_list != self.par_list():
+                        self.messageBox.value = 'Parameters changed! Calculate the raw-image stack first!'
+                    else:
+                        try:
+                            self._viewer.add_image(self.sim.aotf_x,
+                                                   scale=(self.pixel_size.value / self.magnification.value,
+                                                          self.pixel_size.value / self.magnification.value),
+                                                   translate=(
+                                                       -self.pixel_size.value / self.magnification.value * self.N.value / 2,
+                                                       -self.pixel_size.value / self.magnification.value * self.N.value / 2),
+                                                   name='OTF perpendicular to x')
+                        except Exception as e:
+                            print(str(e))
+
+        show_illumination = PushButton(text='illumination')
+
+        @show_illumination.clicked.connect
+        def on_show_illumination_click():
+            if hasattr(self, 'sim'):
+                if self.used_par_list != self.par_list():
+                    self.messageBox.value = 'Parameters changed! Calculate the raw-image stack first!'
+                else:
+                    try:
+                        self._viewer.add_image(self.sim.illumination_stack(),
+                                               scale=(self.zdrift.value * 0.001,
+                                                      self.pixel_size.value / self.magnification.value,
+                                                      self.pixel_size.value / self.magnification.value),
+                                               translate=(-self.zdrift.value * 0.001 * self.tpoints.value / 2,
+                                                          -self.pixel_size.value / self.magnification.value * self.N.value / 2,
+                                                          -self.pixel_size.value / self.magnification.value * self.N.value / 2),
+                                               interpolation2d='spline36',
+                                               name='illumination')
+                    except Exception as e:
+                        print(str(e))
+
+        w_save_and_print = Container(widgets=[Container(widgets=[save_tif_with_tags, show_psf, show_otf]),
+                                              Container(widgets=[print_tif, show_raw_img_sum, show_illumination])],
+                                     layout='horizontal', labels=None)
+        self.messageBox = LineEdit(value='Messages')
+        self.w = Container(widgets=[magicgui(self.select_layer, call_button='Select sample layer'),
+                                    w_parameters, w_cal, w_save_and_print, self.messageBox],
                            labels=None)
