@@ -5,12 +5,12 @@ Some calculations are adapted by the work by Mark Neil @Imperial College London
 __author__ = "Meizhu Liang @Imperial College London"
 
 import numpy as np
+import opt_einsum as oe
 import time
 import tifffile
 
 try:
     import cupy as cp
-
     print('cupy imported')
     import_cp = True
 except:
@@ -42,8 +42,8 @@ class Base_simulator:
     ill_NA = 0.75  # numerical aperture at illumination beams
     det_NA = 1.1  # numerical aperture at sample
     n = 1.33  # refractive index at sample
-    ill_wavelength = 520  # illumination wavelength in nm
-    det_wavelength = 570  # detection wavelength in nm
+    ill_wavelength = 520e-3  # illumination wavelength in um
+    det_wavelength = 570e-3  # detection wavelength in um
     zrange = 7.0  # distance either side of focus to calculate, in microns, could be arbitrary
     dz = 0.4  # step size in axial direction of PSF
     fwhmz = 3.0  # FWHM of light sheet in z
@@ -52,15 +52,14 @@ class Base_simulator:
     defocus = 0  # de-focus aberration in um
     spherical = 0
     sph_abb = 0
-    add_error = True
+    add_error = False
 
     def initialise(self):
         if self.acc == 3:
             self.xp = cp
         np.random.seed(self.random_seed)
         # self.seed(1234)  # set random number generator seed
-        self.ill_wavelength = self.ill_wavelength * 1e-3
-        self.det_wavelength = self.det_wavelength * 1e-3
+        self.ph = 4 * np.pi * self.ill_NA / self.ill_wavelength
         self.sigmaz = self.fwhmz / 2.355
         self.dx = self.pixel_size / self.magnification  # Sampling in lateral plane at the sample in um
         self.dxn = self.det_wavelength / (4 * self.det_NA)  # 2 * Nyquist frequency in x and y.
@@ -69,7 +68,7 @@ class Base_simulator:
         self.res = self.det_wavelength / (2 * self.det_NA)
         oversampling = self.res / self.dxn  # factor by which pupil plane oversamples the coherent psf data
         self.dk = oversampling / (self.Nn / 2)  # Pupil plane sampling
-        self.k0 = 2 * np.pi * self.n / self.det_wavelength
+        self.k0 = 2 * np.pi * self.n / (self.ill_wavelength)
         self.kx, self.ky = self.xp.meshgrid(self.xp.linspace(-self.dk * self.Nn / 2, self.dk * self.Nn / 2 - self.dk, self.Nn),
                                        self.xp.linspace(-self.dk * self.Nn / 2, self.dk * self.Nn / 2 - self.dk, self.Nn))
         self.kr = np.sqrt(self.kx ** 2 + self.ky ** 2)  # Raw pupil function, pupil defined over circle of radius 1.
@@ -94,142 +93,66 @@ class Base_simulator:
         if (self.acc == 1) | (self.acc == 2):
             self._tdev = torch.device('cuda' if self.acc == 2 else 'cpu')
 
-    # def point_cloud_fil(self):
-    #     """
-    #     Generates a point-cloud as the object in the imaging system.
-    #     """
-    #
-    #     L = 5  # full length of each filament in um
-    #     dL = 0.05  # target length of each short bit to form a filament
-    #
-    #     for i in range(self.nSamples):
-    #         alpha = 2 * np.pi * np.random.rand()
-    #         beta = np.pi / 4 * (1 - 2 * np.random.rand())
-    #
-    #         # centres of filaments
-    #         xS = -L / 2 + L * np.random.rand()
-    #         yS = -L / 2 + L * np.random.rand()
-    #         zS = 0
-    #
-    #         # starting points of filaments
-    #         x1 = xS + (L / 2 * np.cos(alpha) * np.cos(beta))
-    #         y1 = yS + (L / 2 * np.sin(alpha) * np.cos(beta))
-    #         z1 = zS + (L / 2 * np.sin(beta))
-    #
-    #         # ending points of filaments
-    #         x2 = xS - (L / 2 * np.cos(alpha) * np.cos(beta))
-    #         y2 = yS - (L / 2 * np.sin(alpha) * np.cos(beta))
-    #         z2 = zS - (L / 2 * np.sin(beta))
-    #
-    #         px = 2 * np.pi * np.random.rand()
-    #         fx = 2 * np.pi * 0.5 * np.random.rand()
-    #         ax = L / 10 * np.random.rand()
-    #         py = 2 * np.pi * np.random.rand()
-    #         fy = 2 * np.pi * 1.0 * np.random.rand()
-    #         ay = L / 10 * np.random.rand()
-    #         pz = 2 * np.pi * np.random.rand()
-    #         fz = 2 * np.pi * 1.0 * np.random.rand()
-    #         az = L / 10 * np.random.rand()
-    #
-    #         # adding some perturbation
-    #         l = np.arange(0, 1, dL / L)
-    #         x = x1 + (x2 - x1) * l + ax * np.cos(fx * l + px)
-    #         y = y1 + (y2 - y1) * l + ay * np.cos(fy * l + py)
-    #         z = z1 + (z2 - z1) * l + az * np.cos(fz * l + pz)
-    #
-    #         xd = x[1:] - x[:-1]
-    #         yd = y[1:] - y[:-1]
-    #         zd = z[1:] - z[:-1]
-    #
-    #         # steps = np.sqrt(xd ** 2 + yd ** 2 + zd ** 2)
-    #         # print(f'length = {np.sum(steps):.2f} um, average = {1000 * np.mean(steps) :.2f} nm, stdev = {1000 * np.std(steps) :.2f} nm')
-    #         if i == 0:
-    #             self.points = np.single(np.dstack([x, y, z]).squeeze())
-    #         else:
-    #             self.points = np.concatenate((self.points, np.single(np.dstack([x, y, z]).squeeze())))
-    #     self.npoints = self.points.shape[0]
-
-    # def point_cloud(self):
-    #     """
-    #     Generates a point-cloud as the object in the imaging system.
-    #     """
-    #     rad = 5  # radius of sphere of points
-    #     # Multiply the points several times to get the enough number
-    #     pointsxn = (2 * np.random.rand(self.nSamples * 3, 3) - 1) * [rad, rad, rad]
-    #
-    #     pointsxnr = np.sum(pointsxn * pointsxn, axis=1)  # multiple times the points
-    #     points_sphere = pointsxn[pointsxnr < (rad ** 2), :]  # simulate spheres from cubes
-    #     self.points = np.single(points_sphere[(range(self.nSamples)), :])
-    #     self.points[:, 2] = self.points[:, 2] / 2  # to make the point cloud for OTF a ellipsoid rather than a sphere
-    #     self.npoints = self.nSamples
-
     def phase_tilts(self):
         """Generates phase tilts in frequency space"""
+
         xyrange = self.Nn / 2 * self.dxn
         dkxy = np.pi / xyrange
         dkz = np.pi / self.zrange
-        if self.acc == 0:
-            self.kxy = np.arange(-self.Nn / 2 * dkxy, (self.Nn / 2) * dkxy, dkxy, dtype=np.single)
-            self.kz = np.arange(-self.Nzn / 2 * dkz, (self.Nzn / 2) * dkz, dkz, dtype=np.single)
-        elif self.acc == 3:
-            self.kxy = cp.arange(-self.Nn / 2 * dkxy, (self.Nn / 2) * dkxy, dkxy, dtype=cp.single)
-            self.kz = cp.arange(-self.Nzn / 2 * dkz, (self.Nzn / 2) * dkz, dkz, dtype=cp.single)
+        if (self.acc == 0) or (self.acc == 3):
+            kxy = self.xp.arange(-self.Nn / 2 * dkxy, (self.Nn / 2) * dkxy, dkxy, dtype=self.xp.single)
+            kz = self.xp.arange(-self.Nzn / 2 * dkz, (self.Nzn / 2) * dkz, dkz, dtype=self.xp.single)
         else:
-            self.kxy = torch.arange(-self.Nn / 2 * dkxy, (self.Nn / 2) * dkxy, dkxy,
+            kxy = torch.arange(-self.Nn / 2 * dkxy, (self.Nn / 2) * dkxy, dkxy,
                                     dtype=torch.float32, device=self._tdev)
-            self.kz = torch.arange(-self.Nzn / 2 * dkz, (self.Nzn / 2) * dkz, dkz,
+            kz = torch.arange(-self.Nzn / 2 * dkz, (self.Nzn / 2) * dkz, dkz,
                                    dtype=torch.float32, device=self._tdev)
 
-        if self.acc == 0:
-            self.phasetilts = np.zeros((self._nsteps, self.Nzn, self.Nn, self.Nn), dtype=np.complex64)
-        elif self.acc == 3:
-            self.phasetilts = cp.zeros((self._nsteps, self.Nzn, self.Nn, self.Nn), dtype=cp.complex64)
+        if (self.acc == 0) or (self.acc == 3):
+            self.phasetilts = self.xp.zeros((self._nsteps, self.Nzn, self.Nn, self.Nn), dtype=self.xp.complex64)
         else:
-            self.phasetilts = torch.zeros((self._nsteps, self.Nzn, self.Nn, self.Nn), dtype=torch.complex64,
+            self.phasetilts = torch.zeros(self._nsteps, self.Nzn, self.Nn, self.Nn, dtype=torch.complex64,
                                           device=self._tdev)
-
         start_time = time.time()
         itcount = 0
         total_its = self._angleStep * self._phaseStep * self.npoints
         lastProg = 0
-        self.ph = 4 * np.pi * self.ill_NA / self.ill_wavelength
+
         for astep in range(self._angleStep):
+            self.jones_vectors(astep)
             for pstep in range(self._phaseStep):
                 self.points += self.drift * np.random.standard_normal(3) / 1000
                 self.points[:, 0] += self.xdrift / 1000
                 self.points[:, 2] += self.zdrift / 1000
                 isteps = pstep + self._phaseStep * astep  # index of the steps
-                for i in range(self.npoints):
-                    prog = (100 * itcount) // total_its
-                    if prog > lastProg + 9:
-                        lastProg = prog
-                        yield f'Phase tilts calculation: {prog:.1f}% done'
-                    itcount += 1
-                    self.x = self.points[i, 0]
-                    self.y = self.points[i, 1]
-                    z = self.points[i, 2]
-                    # get illumination from the child class
-                    if self.pol == 'axial':
-                        ill = self._illAx(pstep, astep)
-                    elif self.pol == 'circular':
-                        ill = self._illCi(pstep, astep)
-                    else:
-                        ill = self._illIp(pstep, astep)
-                    if self.acc == 0:
-                        px = np.exp(1j * np.single(self.x * self.kxy))
-                        py = np.exp(1j * np.single(self.y * self.kxy))[:, np.newaxis]
-                        pz = (np.exp(1j * np.single(z * self.kz)) * ill)[:, np.newaxis, np.newaxis]
-                        self.phasetilts[isteps, :, :, :] += (px * py) * pz
-                    elif self.acc == 3:
-                        px = cp.exp(1j * self.x * self.kxy)
-                        py = cp.exp(1j * self.y * self.kxy)[:, cp.newaxis]
-                        pz = (cp.exp(1j * z * self.kz) * ill)[:, cp.newaxis, cp.newaxis]
-                        self.phasetilts[isteps, :, :, :] += (px * py) * pz
-                    else:
-                        px = torch.exp(1j * self.x * self.kxy)
-                        py = torch.exp(1j * self.y * self.kxy)
-                        pz = torch.exp(1j * z * self.kz) * ill
-                        self.phasetilts[isteps, :, :, :] += (px * py[..., None]) * pz[..., None, None]
+                prog = (100 * itcount) // total_its
+                if prog > lastProg + 9:
+                    lastProg = prog
+                    yield f'Phase tilts calculation: {prog:.1f}% done'
+                itcount += 1
+
+                # get illumination from the child class
+
+                if (self.acc == 0) or (self.acc == 3):
+                    x = self.xp.array(self.points[:, 0], dtype=self.xp.single)
+                    y = self.xp.array(self.points[:, 1], dtype=self.xp.single)
+                    z = self.xp.array(self.points[:, 2], dtype=self.xp.single)
+                    ill = self.xp.array(self._ill_test(x, y, pstep, astep),
+                                        dtype=self.xp.single)
+                    px = self.xp.exp(1j * kxy[self.xp.newaxis, :] * x[:, self.xp.newaxis])
+                    py = self.xp.exp(1j * kxy[self.xp.newaxis, :] * y[:, self.xp.newaxis])
+                    pz = self.xp.exp(1j * kz[self.xp.newaxis, :] * z[:, self.xp.newaxis])
+                    self.phasetilts[isteps, :, :, :] = oe.contract('i,il,ik,ij->jkl', ill, px, py, pz)
+                else:
+                    x = torch.tensor(self.points[:, 0], dtype=torch.float32, device=self._tdev)
+                    y = torch.tensor(self.points[:, 1], dtype=torch.float32, device=self._tdev)
+                    z = torch.tensor(self.points[:, 2], dtype=torch.float32, device=self._tdev)
+                    ill = torch.tensor(self._ill_test(self.points[:,0], self.points[:,1], pstep, astep),
+                                       dtype=torch.float32, device=self._tdev)
+                    px = torch.exp(1j * kxy[None, :] * x[:, None])
+                    py = torch.exp(1j * kxy[None, :] * y[:, None])
+                    pz = torch.exp(1j * kz[None, :] * z[:, None])
+                    self.phasetilts[isteps, :, :, :] = oe.contract('i,il,ik,ij->jkl', ill, px, py, pz)
         self.elapsed_time = time.time() - start_time
         yield f'Phase tilts calculation time:  {self.elapsed_time:3f}s'
 
@@ -323,9 +246,9 @@ class Base_simulator:
                                                             abs(p[i, 0] * Exy + p[i, 1] * Eyy + p[i, 2] * Ezy) ** 2)
                 intensityz = intensityz + excitation3[i] * (abs(p[i, 0] * Exx + p[i, 1] * Eyx + p[i, 2] * Ezx) ** 2 +
                                                            abs(p[i, 0] * Exy + p[i, 1] * Eyy + p[i, 2] * Ezy) ** 2)
-            if self.pol == 'axial':
+            if self.pol == 'r':
                 intensity = intensityz  # for axially polarised illumination
-            elif self.pol == 'circular':
+            elif self.pol == 'c':
                 # dipoles that re-orient between excitation and emmission and maybe for circular polarised illumination
                 intensity = (intensityx + intensityy + intensityz) / 3
             else:
@@ -356,89 +279,6 @@ class Base_simulator:
         psf = psf * self.Nn ** 2 / self.xp.sum(pupil) * self.Nz / self.Nzn
         return psf
 
-    # def raw_image_stack(self):
-    #     # Calculates point cloud, phase tilts, 3d psf and otf before the image stack.
-    #     self.initialise()
-    #     self.drift = 0.0  # no random walk using this method
-    #     self.point_cloud()
-    #     yield "Point cloud calculated"
-    #
-    #     self._nsteps = self._phaseStep * self._angleStep
-    #     for msg in self.phase_tilts():
-    #         yield msg
-    #     if self.psf_calc == 'vector':
-    #         psf = self.get_vector_psf()
-    #     else:
-    #         psf = self.get_scalar_psf()
-    #     self.psf_z0 = psf[int(self.Nzn / 2 + 5), :, :]  # psf at z=0
-    #     if self.acc == 3:
-    #         self.psf_z0 = cp.asnumpy(self.psf_z0)
-    #     yield "psf calculated"
-    #
-    #     # Calculating 3d otf
-    #     otf = self.xp.fft.fftn(psf)
-    #     aotf = abs(self.xp.fft.fftshift(otf))  # absolute otf
-    #     if self.acc == 3:
-    #         aotf = cp.asnumpy(aotf)
-    #     m = max(aotf.flatten())
-    #     aotf_z = []
-    #     if self.acc == 3:
-    #         aotf = cp.array(aotf)
-    #     for x in range(self.Nzn):
-    #         aotf_z.append(self.xp.sum(aotf[x]))
-    #     self.aotf_x = self.xp.log(
-    #         aotf[:, int(self.Nn / 2), :].squeeze() + 0.0001)  # cross section perpendicular to x axis
-    #     if self.acc == 3:
-    #         self.aotf_x = cp.asnumpy(self.aotf_x)
-    #     # aotf_x is the same as aotf_y
-    #     # self.aotf_y = self.xp.log(aotf[:, :, int(self.Nn / 2)].squeeze() + 0.0001)
-    #     yield "3d otf calculated"
-    #
-    #     if (self.acc == 0) | (self.acc == 3):
-    #         img = self.xp.zeros((self.Nz * self._nsteps, self.N, self.N), dtype=self.xp.single)
-    #     else:
-    #         img = torch.empty((self.Nz * self._nsteps, self.N, self.N), dtype=torch.float, device=self._tdev)
-    #
-    #     for i in range(self._nsteps):
-    #         if (self.acc == 0) | (self.acc == 3):
-    #             ootf = self.xp.fft.fftshift(otf) * self.phasetilts[i, :, :, :]
-    #             img[self.xp.arange(i, self.Nz * self._nsteps, self._nsteps), :, :] = self.xp.abs(
-    #                 self.xp.fft.ifftn(ootf, (self.Nz, self.N, self.N)))
-    #         else:
-    #             ootf = torch.fft.fftshift(torch.as_tensor(otf, device=self._tdev), ) * self.phasetilts[i, :, :, :]
-    #             img[torch.arange(i, self.Nz * self._nsteps, self._nsteps), :, :] = (torch.abs(
-    #                 torch.fft.ifftn(ootf, (self.Nz, self.N, self.N)))).to(torch.float)
-    #     # OK to use abs here as signal should be all positive.
-    #     # Abs is required as the result will be complex as the fourier plane cannot be shifted back to zero when oversampling.
-    #     # But should reduction in sampling be allowed here(Nz < Nzn)?
-    #
-    #     stackfilename = f"Raw_img_stack_{self.N}_{self.pol}.tif"
-    #     if self.acc == 0:
-    #         # raw image stack
-    #         # raw image sum along z axis
-    #         self.img_sum_z = np.sum(img, axis=0)
-    #         # raw image sum along x (or y) axis
-    #         self.img_sum_x = np.sum(img, axis=1)
-    #         self.img = img
-    #     elif self.acc == 1:
-    #         self.img_sum_z = (torch.sum(img, axis=0)).numpy()
-    #         self.img_sum_x = (torch.sum(img, axis=1)).numpy()
-    #         self.img = img.numpy()
-    #     elif self.acc == 2:
-    #         self.img_sum_z = (torch.sum(img, axis=0)).detach().cpu().numpy()
-    #         self.img_sum_x = (torch.sum(img, axis=1)).detach().cpu().numpy()
-    #         self.img = img.detach().cpu().numpy()
-    #     elif self.acc == 3:
-    #         self.img_sum_z = cp.asnumpy(cp.sum(img, axis=0))
-    #         self.img_sum_x = cp.asnumpy(cp.sum(img, axis=1))
-    #         self.img = cp.asnumpy(img)
-    #
-    #     # Save generated images
-    #     tifffile.imwrite(stackfilename, self.img)
-    #     yield "file saved"
-    #
-    #     yield f'Finished, Phase tilts calculation time:  {self.elapsed_time:3f}s'
-
     def raw_image_stack_brownian(self):
         # Calculates 3d psf and otf before the image stack
         self.initialise()
@@ -472,7 +312,6 @@ class Base_simulator:
         # self.aotf_y = self.xp.log(aotf[:, :, int(self.Nn / 2)].squeeze() + 0.0001)
         yield "3d otf calculated"
 
-        self._nsteps = self._phaseStep * self._angleStep
         self.points[:, 0] -= self.xdrift * self.tpoints / 2000
         self.points[:, 2] -= self.zdrift * self.tpoints / 2000
         if (self.acc == 0) | (self.acc == 3):
@@ -527,3 +366,35 @@ class Base_simulator:
         tifffile.imwrite(stackfilename, self.img)
         elapsed_Brownian = time.time() - start_Brownian
         yield f'Finished, Phase tilts calculation time:  {elapsed_Brownian:3f}s'
+
+    def illumination_stack(self):
+        # Calculates 3d psf and otf before the image stack
+        print(f'allocating illumination stack: {(int(self.tpoints), self.N, self.N)}')
+
+        illumination = self.xp.zeros((int(self.tpoints), self.N, self.N))
+        xyvals = (self.xp.arange(self.N) - self.N / 2) * self.dx
+        xarr, yarr = self.xp.meshgrid(xyvals, xyvals)
+        xarr_l, yarr_l = self.xp.array(xarr).flatten(), self.xp.array(yarr).flatten()
+        self.npoints = xarr_l.shape[0]
+
+        start_time = time.time()
+        itcount = -1
+        for astep in range(self._angleStep):
+            self.jones_vectors(astep)
+            for pstep in range(self._phaseStep):
+                itcount += 1
+                ill_1d = self._ill_test(xarr_l, yarr_l, pstep, astep)
+                illumination[itcount, :, :] = self.xp.reshape(ill_1d, (self.N, self.N))
+
+        for i in range(1, int(self.tpoints) // self._nsteps):
+            for j in range(self._nsteps):
+                illumination[i * self._nsteps + j, :, :] = illumination[j, :, :]
+
+        elapsed_time = time.time() - start_time
+        print(f'Illumination calculation time:  {elapsed_time:3f}s')
+        if self.acc == 3:
+            return cp.asnumpy(illumination)
+        else:
+            return illumination
+
+
